# coding:utf-8
import torch
import numpy as np
import os
import glob
from os.path import join
import random
import time


# 参考代码：adaptive span
##############################################################################
# CHECKPOINT
##############################################################################
def load_checkpoint(checkpoint_path, model, optimizer, args):
    print(f'loading from a checkpoint at {checkpoint_path}')
    checkpoint_state = torch.load(checkpoint_path)
    log_file = checkpoint_state['log_file']
    epoch_num = checkpoint_state['epoch_num']  # next iteration
    model.load_state_dict(checkpoint_state['model'])
    optimizer.load_state_dict(checkpoint_state['optimizer'])
    best_val = checkpoint_state['best_val']
    best_epoch_num = checkpoint_state['best_epoch_num']
    early_stop_acc = checkpoint_state['early_stop_acc']
    accumulate_time = checkpoint_state['accumulate_time']

    return log_file, epoch_num + 1, best_val, best_epoch_num, early_stop_acc, accumulate_time


def save_checkpoint(checkpoint_path, log_file, epoch_num, model, optimizer, best_val, best_epoch_num, early_stop_acc,
                    accumulate_time):
    if checkpoint_path:
        checkpoint_state = {
            'log_file': log_file,
            'epoch_num': epoch_num,  # last completed iteration
            # todo 注意这里可能需要是model.module.state_dict()，中断一下看是否能加载
            'model': model.state_dict(),
            'optimizer': optimizer.state_dict(),
            'best_val': best_val,
            'best_epoch_num': best_epoch_num,
            'early_stop_acc': early_stop_acc,
            'accumulate_time': accumulate_time
        }

        torch.save(checkpoint_state, checkpoint_path)


def log_the_parameters(args, log_file):
    with open(log_file, 'a') as result_file:
        result_file.write(f'hyper parameters: ')
        # argparse打印参数见：https://stackoverflow.com/questions/27181084/how-to-iterate-over-arguments
        for key, val in vars(args).items():
            result_file.write(f'{key}:{val}, ')
        result_file.write('\n')


def logger(text, log_file):
    with open(log_file, 'a') as result_file:
        result_file.write(f'{text}\n')


def logger_pure(text, log_file):
    with open(log_file, 'a') as result_file:
        result_file.write(text)


def get_curr_data_name(args):
    # 获取命令行里数据集路径里的文件名
    log_base = os.path.basename(args.data)
    # 把文件名的扩展去掉
    log_file_no_ext = os.path.splitext(log_base)[0]

    return log_file_no_ext


def get_curr_log_name(args):
    # log directory check, 参考：https://thispointer.com/how-to-create-a-directory-in-python/
    if not os.path.exists(args.log_dir):
        os.mkdir(args.log_dir)

    # 获取命令行里数据集路径里的文件名
    log_base = os.path.basename(args.data)
    # 把文件名的扩展去掉
    log_file_no_ext = os.path.splitext(log_base)[0]
    # 创建新的相关数据集的log文件
    curr_log_files = glob.glob(join(args.log_dir, log_file_no_ext + '_res_' + '*.txt'))
    if len(curr_log_files) != 0:
        file_names = [os.path.basename(name) for name in curr_log_files]
        file_names_no_ext = [os.path.splitext(name)[0] for name in file_names]
        # 用split而不用partition的原因见：https://www.quora.com/Whats-the-difference-between-partition-and-split-for-strings-in-Python
        nums = [int(name.split('_')[-1]) for name in file_names_no_ext]
        nums.sort()
        max = nums[-1]
        log_file = join(args.log_dir, f'{log_file_no_ext}_res_{max + 1}.txt')
    else:
        log_file = join(args.log_dir, f'{log_file_no_ext}_res_0.txt')

    return log_file


def get_curr_loss_name(args):
    # log directory check, 参考：https://thispointer.com/how-to-create-a-directory-in-python/
    if not os.path.exists(args.log_dir):
        os.mkdir(args.log_dir)

    # 获取命令行里数据集路径里的文件名
    log_base = os.path.basename(args.data)
    # 把文件名的扩展去掉
    log_file_no_ext = os.path.splitext(log_base)[0]
    # 创建新的相关数据集的log文件
    curr_log_files = glob.glob(join(args.log_dir, log_file_no_ext + '_loss_' + '*.txt'))
    if len(curr_log_files) != 0:
        file_names = [os.path.basename(name) for name in curr_log_files]
        file_names_no_ext = [os.path.splitext(name)[0] for name in file_names]
        # 用split而不用partition的原因见：https://www.quora.com/Whats-the-difference-between-partition-and-split-for-strings-in-Python
        nums = [int(name.split('_')[-1]) for name in file_names_no_ext]
        nums.sort()
        max = nums[-1]
        log_file = join(args.log_dir, f'{log_file_no_ext}_loss_{max + 1}.txt')
    else:
        log_file = join(args.log_dir, f'{log_file_no_ext}_loss_0.txt')

    return log_file


def get_curr_best_check_path(args):
    # 获取命令行里数据集路径里的文件名
    log_base = os.path.basename(args.data)
    # 把文件名的扩展去掉
    log_file_no_ext = os.path.splitext(log_base)[0]
    path_name = join(args.save, f'{log_file_no_ext}_{args.horizon}_step.pt')

    return path_name


def get_curr_checkpoint_path(args):
    # 获取命令行里数据集路径里的文件名
    log_base = os.path.basename(args.data)
    # 把文件名的扩展去掉
    log_file_no_ext = os.path.splitext(log_base)[0]
    path_name = join(args.checkpoints_dir, f'{log_file_no_ext}_{args.horizon}_step.pt')

    return path_name


def count_parameters(model):
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


def set_rand_seed(args):
    # 随机生成种子并保存在args里面，这样就不用每次实验都手动设置不同的种子的
    seed = random.randint(1, 1000000)
    args.seed = seed
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)


def adjust_learning_rate(optimizer, epoch, lr_decay=0.3, lr_decay_epoch=5):
    """Decay learning rate by a factor of lr_decay every lr_decay_epoch epochs"""
    if epoch % lr_decay_epoch:
        return optimizer

    for param_group in optimizer.param_groups:
        param_group['lr'] *= lr_decay

    return optimizer
